# -*- coding: utf-8 -*-

'''
This module implements classes to extend the browser with Praat scripts. 
'''

import os, sys 
import re
from glob import glob
import locale, codecs

import wx
import wx.lib.filebrowsebutton as filebrowse
from messages import *
import pfc.standard
from zipfile import ZipFile, ZIP_DEFLATED, is_zipfile

if __name__ == '__main__':
	sys.path.append('..')
from core.regex import Pattern


class PraatConsoleError(Exception):
	pass

class BufferPathError(Exception):
	pass

class PluginMainError(Exception):
	"Raised when no main() is found."
	pass

class BufferNotFound(Exception):
	pass

_fbnames = \
{
	'folder': 'Dossier',
	'textgrid': 'TextGrid',
	'wav': 'Fichier WAV'
}

def _toTitle(s):
	# only first letter, even in names such as my_var
	s = s[0].upper() + s[1:]
	return s.replace('_', ' ')

def _getPyFromPraat(script):
	"Get a /path/to/foo.py from /path/to/foo.praat"
	root = os.path.dirname(script)
	wav = os.path.basename(script)
	
	base, ext = os.path.splitext(script)
	tgd = base + '.py'
	
	return os.path.join(root, tgd)	

def _getPraatFromPlugin(plugin):
	praatscript = '%s.praat' % os.path.splitext(os.path.basename(plugin))[0]
	return os.path.join(plugin, praatscript)

#---------------------------------------------------------------------------------------------

def importScript(prefsdir):
	oldscript = ''
	scriptsdir = os.path.join(prefsdir, 'scripts')
	
	# Get file path	
	dialog = wx.FileDialog(None, u"Script à importer", wildcard = "Script Praat (*.praat)|*.praat|")
	
	if dialog.ShowModal() == wx.ID_OK:
		oldscript = dialog.GetPath()
		
	dialog.Destroy()
	if not oldscript:
		return

	if not os.path.exists(scriptsdir):
		os.mkdir(scriptsdir)
		
	newscript = os.path.join(scriptsdir, os.path.basename(oldscript))
	old = open(oldscript)
	new = open(newscript, 'w')
	new.write(old.read())
	old.close()
	new.close()
	
	info(u"Le fichier a été importé avec succès.")


def importPlugin(prefsdir):
	plugin = ''
	scriptsdir = os.path.join(prefsdir, 'scripts')
	
	dialog = wx.FileDialog(None, u"Script à importer", wildcard = "Fichier ZIP (*.zip)|*.zip|")	
	if dialog.ShowModal() == wx.ID_OK:
		plugin = dialog.GetPath()
		
	dialog.Destroy()
	if not plugin:
		return
	
	if not is_zipfile(plugin):
		alert(u"Le fichier ne semble pas être un fichier ZIP correct.")
		return

	if not os.path.exists(scriptsdir):
		os.mkdir(scriptsdir)
		
	for dir in unzip(plugin):
		newdir = os.path.join(scriptsdir, dir)
		try:	
			os.rename(dir, newdir)
		except:
			alert(u"Erreur lors de l'import du greffon.")
		
	info(u"Le greffon a été importé avec succès.")
	
	
def is_plugin(dir):
	"Test whether a path is a (valid) PFC plugin."
	if not (dir.endswith('.plugin') and os.path.isdir(dir)):
		return False
	
	base = os.path.basename(dir).split('.')[0]
	pythonScriptFound = False
	praatScriptFound  = False
	
	for file in os.listdir(dir):
		if file == '%s.py' % base:
			pythonScriptFound = True
		elif file == '%s.praat' % base:
			praatScriptFound = True
			
	if praatScriptFound and pythonScriptFound:
		return True
	else:
		return False
	
	
def unzip(archive):
	"Unzip a plugin and return a list of directories."
	zip = ZipFile(archive, 'r', ZIP_DEFLATED)
	names = zip.namelist()
	dirs = [ ] # list of directories to move
	for name in names:
		if not os.path.exists(os.path.dirname(name)):
			os.mkdir(os.path.dirname(name))
			# append if it is not a subdirectory
			append = True
			for dir in dirs:
				if name.startswith(dir):
					append = False
			if append:
				dirs.append(name)
		try:
			buffer = open(name, 'wb') 
			buffer.write(zip.read(name))
			buffer.close()
		except IOError:
			pass	
	zip.close()

	return dirs
		
#---------------------------------------------------------------------------------------------

def getScripts(folders):
	'Return all scripts and plugins from a list of folders.'
	scripts = [ ]
	
	for folder in folders:
		if not os.path.exists(folder):
			continue
		map(lambda s: scripts.append(s), glob("%s%s*.praat"  % (folder, os.sep)))
		map(lambda s: scripts.append(s), glob("%s%s*.plugin" % (folder, os.sep)))

	return scripts
	
	
def getIncludeScripts(scripts):
	'''
	Return all include scripts and plugins from a list of scripts. 
	Attachable scripts have special comments of the kind:

	## if the following value is True, the script is attached to the Scripts menus.
	# $INCLUDE = True
	## label of the script in the script menu
	# $LABEL = "Vérifier les frontières"
	## wx.StaticText to attach to the ScriptFrame
	# $SHORT_DESCRIPTION = "Script pour vérifier les frontières"
	## File extension to look for when calling the script (should be 'TextGrid' or 'wav')
	# $EXTENSION = "TextGrid"
	## self-explanatory...
	# $ACCESS_KEY = F
	## buffer file as a Praat string variable.
	# $BUFFER = var$
	
	The two compulsory values are $INCLUDE and $LABEL. $BUFFER stores a Praat variable 
	(here var$) which contains a buffer file. The script looks for a line like:
	var$ = "filename"
	and returns filename. This allows to send back data to the program.
	
	A plugin is a directory "foo.plugin" containing at list "foo.praat" and "foo.py".
	'''	
	includePtn = re.compile('\$INCLUDE\s+=\s+"?True"?')
	labelPtn   = Pattern('\$LABEL\s+=\s+"(.+)"')
	shortPtn   = Pattern('\$SHORT_DESCRIPTION\s+=\s+"(.+)"')
	extPtn     = Pattern('\$EXTENSION\s+=\s+"(.+)"')
	accessPtn  = Pattern('\$ACCESS_KEY\s+=\s+"?([a-zA-Z])"?')
	bufferPtn  = Pattern('\$BUFFER\s+=\s+(.*\$)')	
	
	includes = { } # stores included scripts as dictionaries {path: {properties}}
	
	for script in scripts:
		include = False
		props = { }
		
		if is_plugin(script):
			sys.path.append(script) # update the PATH with plugin directories
			props['plugin'] = script
			script = _getPraatFromPlugin(script)

		infile = codecs.open(script, 'r', 'utf-16')	
		lines = infile.readlines()
		infile.close()
		
		for line in lines:
			if includePtn.search(line):
				include = True
				
			elif labelPtn.search(line):
				props['label'] = labelPtn.match #ICI
				
			elif shortPtn.search(line):
				props['short_desc'] = shortPtn.match
				
			elif extPtn.search(line):
				props['extension'] = extPtn.match
			
			elif accessPtn.search(line):
				props['access'] = accessPtn.match
				
			elif bufferPtn.search(line):
				props['buffer'] = bufferPtn.match
		
		# look for the buffer path	
		if 'buffer' in props:
			ptn = Pattern('%s\s+=\s+"(.+)"' % props['buffer'].replace('$', '\$'))
			
			for line in lines:
				if ptn.search(line):
					props['buffer_path'] = ptn.match
					
			if 'buffer_path' in props:
				# clean up
				props['buffer'] = props['buffer_path']
				del props['buffer_path']
			else:
				raise BufferPathError, "The variable %s was not found in file %s" \
						% (props['buffer'], script)	
						
						
		# Get the form
		props['form'] = _parseForm(lines)
						
		if include:
			includes[script] = props	
	
	return includes	
	
		
def _parseForm(lines):
	'Parse the form in a Praat script. The script must be sent as a list of lines.'
	process = False
	form = [ ] # store each line in the form as a dictionary.
	# comment => wx.StaticText
	comment = Pattern("comment\s+(.+)")
	# this special field is translated to a dir selector. There are 3 cases handled:
	# (1) there is only ONE textgrid field -> $EXTENSION must be "TextGrid" (default)
	# (2) there is only ONE wav field -> $EXTENSION must be "wav"
	# (3) there is ONE TextGrid and ONE wav IN THIS ORDER -> $EXTENSION = "TextGrid|wav"
	# All other cases are not handled and will *break*. If a field is not recognised as a
	# chooser, it will be considered as a regular entry.
	# update 22/10/07: now handles flac files as well as  wav files.
	# TODO: handle more formats in plugins
	chooser = Pattern("sentence\s+(textgrid|wav|flac)")
	# text and numbers are translated to wx.TextCtrl objects
	entry   = Pattern("(?:positive|integer|natural|word|sentence|text)\s+(.+?)(?:\s|$)") # ignore default
	# choice
	radiobox = Pattern("choice\s+(.+?)(?:\s|:|$)") # ignore default value
	option   = Pattern("button\s+(.+)")
	
	# checkbox
	checkbox = Pattern("boolean\s+(.+)")
	
	for i in range(len(lines)):
		line = lines[i]
		# look for beginning of form
		if line.find('form') == 0:
			process = True
			
		if not process:
			continue

		# scan each line
		if comment.search(line):
			form.append({'object': 'comment', 'text': comment.match})
		
		elif chooser.search(line):
			form.append({'object': 'chooser', 'type': chooser.match})
			
		elif entry.search(line):
			form.append({'object': 'entry', 'variable': _toTitle(entry.match)})

		elif radiobox.search(line):
			props = {'object': 'radiobox', 'text': radiobox.match, 'options': [ ]}
			# look for options in next lines
			j = i + 1
			while option.search(lines[j]):
				props['options'].append(option.match)
				j += 1		
			form.append(props)
		
		elif checkbox.search(line):
			form.append({'object': 'checkbox', 'variable': _toTitle(checkbox.match)})
	
		# end of form: stop parsing
		elif line.find('endform') == 0:
			break
		
	return form


#------------------------------------------------------------------------------------------
class Plugin:
	"""
	This class creates a dynamically loaded module. A plugin MUST define a main() function.
	Example of a basic plugin:

	# dummy.py: a dummy script to print the content of the buffer	
	from gui import plugins

	def main():
		buffer = open(plugins.getBuffer())
	
		for line in buffer.readlines():
			print line
		
		buffer.close()
	"""
	# stored here and not in gui.plugins because users shouldn't instanciate it.
	def __init__(self, script):
		import imp
		
		self.path = script
		
		base, ext = os.path.splitext(os.path.basename(script))
		self.module = imp.new_module(base)
		code = open(script)
		
		exec code in self.module.__dict__
		
	def execute(self):
		if not hasattr(self.module, "main"):
			alert("The plug-in %s has no main() function." % self.path)
			return
		
		self.module.main()

class ScriptForm(wx.Frame):
	"Generic frame"
	def __init__(self, parent, scriptpath, script, praat):
		self.parent = parent
		self.praat = praat
		self.script = script
		self.scriptpath = scriptpath
		
		# if the secondary extension, the program will automatically look for the corresponding
		# wav file. For example, if the file is /path/to/11agm1tg.TextGrid, it'll look for
		# /path/to/11agm1tw.wav
		if not 'extension' in self.script:
			self.extension = 'TextGrid'
		elif '|' in self.script['extension']:
			self.extension, self.secondExtension = self.script['extension'].split('|')
		else:
			self.extension = self.script['extension']

		self.SetGUI()				
		self.Show(True)
	
	def OnQuit(self, e = None):
		self.Destroy()
		
				
	def Execute(self, e = None):
		"Execute script by calling Process()."
		msg = u"Le script a été exécuté avec succès."
		args = [ ]
		for getChoice, control in self.callbacks:
			args.append(getChoice(control))
			
		try:	
			self.Process(self.root, args)
		except PraatConsoleError:
			alert(u"Une erreur s'est produite dans l'appel à Praat.")
			return
		
		# Get the buffer path
		if 'buffer' in self.script:
			buffer = self.script['buffer']
			
			if not os.path.exists(buffer):
				# look in the app's scripts dir
				if os.path.exists(os.path.join('scripts', buffer)):
					buffer = os.path.join('scripts', buffer) 
				# look in prefs scripts dir
				elif os.path.exists(os.path.join(self.parent.prefsdir, 'scripts', buffer)):
					buffer = os.path.join(self.parent.prefsdir, 'scripts', buffer)
				# look if it's a plugin
				elif 'plugin' in self.script:
					buffer = os.path.join(self.script['plugin'], buffer)
				# give up
				else:
					raise BufferNotFound, "The buffer %s was not found" % buffer	
		
		# look if a python module exists and load it
		pyscript = _getPyFromPraat(self.scriptpath)
		if os.path.exists(pyscript):
			if 'buffer' in self.script:
				# the new buffer will be available via gui.plugins.getBuffer()
				newbuffer =  os.path.join(self.parent.prefsdir, 'buffer')
				if os.path.exists(newbuffer):
					os.remove(newbuffer)
				os.rename(buffer, newbuffer)

			plugin = Plugin(pyscript)
			plugin.execute()
			# clean up
			os.remove(newbuffer)
			
		# if the buffer exists but there's no plugin, just inform the user that some data has
		# been written
		else:
			if 'buffer' in self.script and os.path.exists(buffer):
				msg += u" Des données ont été écrites dans le fichier %s." % self.script['buffer']
		
			info(msg) # this is not displayed if a python plugin is executed
			
		self.OnQuit()


	def Process(self, path, args):
		"Scan recursively folder to call RunScript() if a file matching the extension is found."
		for f in os.listdir(path):
			if f.startswith('.'):
				continue		
			dir = os.path.join(path, f)
			if os.path.isdir(dir):
				self.Process(dir, args)
			elif dir.lower().endswith(self.extension.lower()): # dir is a TextGrid, flac, wav
				self.RunPraatScript([self.praat, self.scriptpath, dir] + args)

		
	def RunPraatScript(self, args):
		"Call praat from the command line to execute the script."
		# if the secondary file is set, insert it in the arg list after the TextGrid file
		if hasattr(self, "secondExtension"):
			wavFile = pfc.standard.getWavFromTextGrid(args[2])
			args.insert(3, wavFile)

		# escape paths which contain spaces
		for i in range(len(args)):
			if i == 0:
				continue
			try:
				args[i] = '"%s"' % args[i]			
			except:
				pass
		cmd = ' '.join(map(str,args))
		###print cmd
		
		error = os.system(cmd)
		if error:
			raise PraatConsoleError
	
				
	def SetGUI(self):
		"Convert praat script form into a wx frame."
		#win = self.parent
		script = self.script
	
		wx.Frame.__init__(self, self.parent, -1, script['label'].replace('.',''), pos = (150, 50), 
				style = wx.DEFAULT_FRAME_STYLE ^ (wx.RESIZE_BORDER | wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX))
		panel = wx.Panel(self, -1)

		sizer = wx.BoxSizer(wx.VERTICAL)
						
		callbacks = [ ] # store callbacks in the order they appear
		chooserSet = False # set only one chooser
		height = 60
		thickness = 30 # height of every line in the sizer
		
		for item in script['form']:
			control = None
			
			if item['object'] == 'comment':
				control = wx.StaticText(panel, -1, item['text'] + ' ')
							
			elif item['object'] == 'chooser' and not chooserSet:
				control = filebrowse.DirBrowseButton(panel, -1, size = (420, -1),
								changeCallback = self.OnSelectDir, buttonText = 'Choisir')
				control.SetLabel(u'Chemin : ')
				self.chooser = control
				chooserSet = True
				
				
			elif item['object'] == 'entry':
				control = wx.BoxSizer()
				
				label = wx.StaticText(panel, -1, "%s : " % item['variable'])
				width = 420 - label.GetSize()[0]
				entry = wx.TextCtrl(panel, -1, "", size = (width, -1))
				control.Add(label)
				control.Add(entry)
				callbacks.append([self.GetEntryParam, entry])
				
			elif item['object'] == 'checkbox':
				control = wx.CheckBox(panel, -1, item['variable'])
				callbacks.append([self.GetCheckBoxParam, control])
				
			elif item['object'] == 'radiobox':
				control = wx.RadioBox(panel, -1, item['text'], wx.DefaultPosition, 
								(420, -1), item['options'], 1, style = wx.RA_SPECIFY_COLS)
				callbacks.append([self.GetRadioBoxParam, control])
				height += len(item['options']) * thickness
				
			if control:
				sizer.Add(control, 0, wx.ALL, 5)
				height += thickness # adjust size
				
		# update size of form
		if height < 105:
			height = 105
		size = (450, height)
		self.SetSize(size)
		panel.SetSize(size)
		
	
		# add standard buttons
		buttons = wx.BoxSizer()
		cancelBtn = wx.Button(panel, -1, "Annuler")
		okBtn     = wx.Button(panel, -1, u"Exécuter")
		buttons.Add(cancelBtn, 0, wx.ALL, 5)
		buttons.Add(okBtn, 0, wx.ALL, 5)
		sizer.Add(buttons)

		self.Bind(wx.EVT_BUTTON, self.OnQuit, cancelBtn)
		self.Bind(wx.EVT_BUTTON, self.Execute, okBtn)
		
		# keep track of callbacks
		self.callbacks = callbacks

		self.SetSizer(sizer)
		
	def OnSelectDir(self, e = None):
		self.root = self.chooser.GetValue()
		
	def GetEntryParam(self, entry):
		return entry.GetValue()
			
	def	GetCheckBoxParam(self, checkbox):
		if checkbox.IsChecked():
			return 1
		else:
			return 0
		
	def GetRadioBoxParam(self, radiobox):
		choice = radiobox.GetSelection()
		return radiobox.GetString(choice)
	

		
#------------------------------------------------------------------------------------------
		
if __name__ == '__main__':
	app = wx.App()
	os.chdir('..')
	s = getScripts(('scripts', os.path.expanduser('~/.PFCPrefs/scripts')))
	inc = getIncludeScripts(s)
	for key in inc.keys():
		#print key, inc[key]['label']
		print inc[key]['form']
	
	app.MainLoop()